#!/usr/bin/env python3

from z3 import *
from collections import defaultdict

def get_text_file_contents_as_list (filename):
    f=open(filename,"r")
    tmp=f.readlines()
    f.close()
    return [item.rstrip() for item in tmp]

songs=get_text_file_contents_as_list("songs.txt")

# k=disk number; v=list of tracks
table=defaultdict(list)

disks_tbl=[]
disk_n=0
f=open("disks.txt", "r")
cur_disk=None
for x in f:
    x=x.rstrip()
    if x.startswith ("* "):
        cur_disk=x[2:]
        disks_tbl.append(cur_disk)
    elif len(x)==0:
        cur_disk=None
        disk_n+=1
    else:
        #print (cur_disk, "-", x)
        #print (cur_disk, "-", songs.index(x))
        table[disk_n].append(songs.index(x))
disks_total=disk_n
print (f"{disks_total=}")

# that will be a list of all tracks in all disks:
all_tracks=set()
# k=track, v=list of disks, which contains that track:
disks_for_track=defaultdict(list)

for disk in table:
    all_tracks|=set(table[disk])
    for track in table[disk]:
        disks_for_track[track].append(disk)

# int variable for each disk:
disks=[Int('disk_%d' % (t)) for t in range(disks_total)]

# this is optimization problem, so Optimize() instead of Solver():
opt=Optimize()

# each disk variable is either 0 (absent) or 1 (present):
for t in disks:
    opt.add(Or(t==0, t==1))

# we know which disks contain each track
# so we enumerate all disks when preparing expression for each track
# we form expression like "disk_1==1 OR disk_2==1 OR ..." for each track:
for track in list(all_tracks):
    expressions=[disks[s]==1 for s in disks_for_track[track]]
    # expression is a list which unfolds as list of arguments into Z3's Or() function (see asterisk)
    # that results in big expression like "Or(disk_1==1, disk_2==1, ...)"
    # the expression is then added as a constraint:
    print (f"track '{songs[track]}':")
    print (Or(*expressions))
    opt.add(Or(*expressions))

# we need to find a such solution, where minimal number of all "disk_X" variables will have 1
# "*disks" unfolds to a list of arguments: [disk_1, disk_2, disk_3,...]
# "Sum(*disks)" unfolds to the following expression: "Sum(disk_1, disk_2, ...)"
# the sum of all "disk_X" variables should be as minimal as possible:
h=opt.minimize(Sum(*disks))

print (opt.check())
m=opt.model()

disks_in_result=0
for d in range(disks_total):
    if m[disks[d]].as_long()==1:
        print ("* ", disks_tbl[d])
        disks_in_result+=1
        #for s in table[d]:
        #    print (songs[s])
print (f"{disks_in_result=}")

